{
    This file is part of CheSkymp,
    Copyright (C) 2018 ChebMaster.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; If not, see http://www.gnu.org/licenses/

 **********************************************************************}




{$coperators on}

unit skymp;

// a *modified* Vampyre 0.26  guaranteed to work on x86-64 and arm:
{$unitpath ./Imaging/Source;./Imaging/Source/ZLib}
{$includepath ./Imaging/Source}


{$mode objfpc}{$H+}

{$overflowchecks off}
{$rangechecks off}
{$S-}
{$if defined(cpu386)}
  {$fputype sse3}
{$endif}
{$if defined(cpux86_64)}
  {$fputype sse64}
{$endif}
{$optimization level3,regvar,asmcse,orderfields,fastmath,uncertain}



interface

uses
  Classes, SysUtils, IniFiles, ImagingTypes, Imaging, math, typinfo;

type
  float = single;

  TLayerMode = (lmNormal, lmMultiply, lmDivide,
    lmScreen, lmOverlay, lmDodge, lmBurn, lmHard_Light,lmSoft_Light, lmGrain_Extract, lmGrain_Merge,
    lmDifference, lmAddition, lmSubstract, lmDarken_Only, lmLighten_Only,
    lmHue, lmSaturation, lmColor, lmValue);
  TMaskMode = (mmValue, mmAlphaOnly, mmColor);

  TFloatColor = array[0..3] of float; // rgba
  PFloatColor = ^TFloatColor;

  TVector3f = array[0..2] of float;

  PPVector4f = ^PVector4f;
  PVector4f = ^TVector4f;
  TVector4f = packed array [0..3] of float;
  PMatrix4f = ^TMatrix4f;
  TMatrix4f = packed array [0..3] of TVector4f;

  TStringArray = array of string;

type
  ESkympDie = class (Exception) end;

  TSkymp = class;
  TAbstractLayer = class;
  CAbstractLayer = class of TAbstractLayer;
  PLayer = ^TAbstractLayer;
  TLayerArray = array of TAbstractLayer;



  TAbstractLayer = class
  protected
    skymp: TSkymp;
    fname: string;
    mask: TLayerArray;
    opacity: float;
    mode: TLayerMode;
    maskMode: TMaskMode;
    function UnprojectVectorToBitmap(const vector: TVector3f; out x, y: float): boolean; virtual; abstract;
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); virtual; abstract;
    function VectorToPixelConeWidth(const vector: TVector3f): float; virtual; abstract;
    procedure SampleCone(next: PLayer; const vector: TVector3f; const coneWidth, pixelWidth: float; out color: TColorFPRec);  virtual;
    function WrapBitmapCoordinates(x, y: integer; out x1, y1: integer): boolean; virtual; abstract;
    function SampleNextAndBlend(const MyColor: TColorFPRec; next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; virtual;
    procedure ParseOpacityDefinition;
  public
    flip: boolean;
    property name: string read fname;
    function Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; virtual; abstract;
    constructor Create(_name: string; _skymp: TSkymp); virtual;
    function Blend(my, next: TColorFPRec): TColorFPRec;
  end;


  TAbstractBitmapLayer = class(TAbstractLayer)
  protected
    matrix, revmat: TMatrix4f;
    image: TImageData;
    procedure ProduceImage; virtual;
    procedure GetInterpolatedPixel(const x, y: float;  out color: TColorFPRec); virtual;
  public
    function Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; override;
    constructor Create(_name: string; _skymp: TSkymp); override;
  end;


  TAbstractOutputLayer = class (TAbstractBitmapLayer)
    filename: string;
    Layers: TLayerArray;
    inpath, outpath, outname, outext: string;
    supersamples: integer;
    equwidth, equheight: float;
    procedure ProduceImage; override;
    procedure ParseDimensions(out w, h: integer); virtual; abstract;
  public
    procedure Iterate; virtual;
    procedure SaveBitmap; virtual;
    constructor Create(_name: string; _skymp: TSkymp); override;
  end;
  CAbstractOutputLayer = class of TAbstractOutputLayer;


  TSpriteLayer = class (TAbstractBitmapLayer)
  protected
    function UnprojectVectorToBitmap(const vector: TVector3f; out x, y: float): boolean; override;
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); override;
    function VectorToPixelConeWidth(const vector: TVector3f): float; override;
    function WrapBitmapCoordinates(x, y: integer; out x1, y1: integer): boolean; override;
  public
    xangle, yangle: float;
    constructor Create(_name: string; _skymp: TSkymp); override;
  end;


  TSphereMapLayer = class (TAbstractBitmapLayer)
  protected
    function UnprojectVectorToBitmap(const vector: TVector3f; out x, y: float): boolean; override;
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); override;
    function VectorToPixelConeWidth(const vector: TVector3f): float; override;
    function WrapBitmapCoordinates(x, y: integer; out x1, y1: integer): boolean; override;
  public
    //if the sphere map is closed (as in case of simpler rendering where the edges
    // of the sphere are joined) then the circumference is one pixel less than
    // image width and pixels on the right and left sides have to match perfectly
    // This is not case with maps painted manually created using The Gimp
    // with layer warp operations to create a cyclic image
    // and probably not the case for maps rendered in Bryce or such.
    NonClosedHorizontally: boolean;
    constructor Create(_name: string; _skymp: TSkymp); override;
  end;

  TLayerGroup = class(TAbstractLayer)
  protected
    Layers: TLayerArray;
  public
    constructor Create(_name: string; _skymp: TSkymp); override;
    function Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; override;
  end;


  TSolidColorLayer = class(TAbstractLayer)
  protected
    MyColor: TColorFPRec;
  public
    constructor Create(_name: string; _skymp: TSkymp); override;
    function Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; override;
  end;


  TGradientLayer = class(TAbstractLayer)
  protected
    Color1, Color2: TColorFPRec;
    angle1, angle2, pow: float;
    matrix: TMatrix4f;
  public
    constructor Create(_name: string; _skymp: TSkymp); override;
    function Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean; override;
  end;


  TOutSphereMapLayer = class (TAbstractOutputLayer)
  protected
    NonClosedHorizontally: boolean;
    procedure ParseDimensions(out w, h: integer); override;
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); override;
    function VectorToPixelConeWidth(const vector: TVector3f): float; override;
  public
    constructor Create(_name: string; _skymp: TSkymp); override;
  end;

  TOutDoomSkyLayer = class (TOutSphereMapLayer)
  protected
    procedure ParseDimensions(out w, h: integer); override;
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); override;
  end;


  TOutCubeMapLayer = class (TAbstractOutputLayer)
  protected
    procedure BitmapToProjectVector(const x, y: float; out vector: TVector3f); override;
    function VectorToPixelConeWidth(const vector: TVector3f): float; override;
    procedure ParseDimensions(out w, h: integer); override;
  public
    num: integer;
    constructor Create(_name: string; _skymp: TSkymp; _num: integer);
  end;

  COutCubeMapLayer = class of TOutCubeMapLayer;


  Tskymp = class (TIniFile)
  protected
    prevprogress: integer;
    procedure OutputLayerFactory(name: string);
  public
    inpath, outpath, outname, outext: string;
    List, OReserved: TStringList;
    Layers, OutLayers: TLayerArray;
    supersamples: integer;
    equwidth, equheight: float;
    is_debug: boolean;
    procedure ProgressCallback(p: float); virtual;
    procedure OutStr(s: string); virtual;
    constructor Create(fname: string);
    function Execute: boolean;
    procedure Die(errormessage: string); virtual;
    function IsRelativePath(path: string): boolean;
    function ParseLayerDefinition(section_name: string; ident: string = 'layers'; are_output: boolean = false): TLayerArray;
    function ParseColorDefinition(section_name: string; ident: string = 'color'): TColorFPRec;
    function LayerFactory(name: string): TAbstractLayer;
    function ParseMode(_name: string): TLayerMode;
    destructor Destroy;
  end;

implementation

  var
    dbgr,
    dbgg,
    dbgb: float;

  function Explode(s: string; NoLowerCase: boolean = false): TStringArray; forward;
  function VyFormatName(i: TImageFormat) : String; forward;
  operator * (a: TVector3f; b: float) c: TVector3f; forward;
  operator * (a: TVector3f; b: TMatrix4f) c: TVector3f; forward;
  operator * (a, b: TMatrix4f) c: TMatrix4f; forward;
  procedure Normalize(var v: TVector3f); forward;
  procedure LoadIdentity(var matrix: TMatrix4f); forward;
  procedure Rotate(var matrix: TMatrix4f; angle: float; const vector: TVector3f); forward;
  procedure Invert(const a: TMatrix4f; out b: TMatrix4f); forward;
  procedure FindHorizontalPerpendicularVector(const v1: TVector3f; out v2: TVector3f); forward;
  function DotProduct(var a, b: TVector3f): float; forward;




  function arcsin(x: float): float; inline; // FffFFffFfuuuuu----
  begin
    Result:= math.arcsin(max(float(-1.0), min(float(1.0), x)));
  end;

  function arccos(x: float): float; inline;
  begin
    Result:= math.arccos(max(float(-1.0), min(float(1.0), x)));
  end;


  destructor TSkymp.Destroy;
  var
    i: integer;
  begin
    List.Free;
    OReserved.Free;
    for i:=0 to High(OutLayers) do OutLayers[i].Free;
    inherited;
  end;

  function TSkymp.ParseLayerDefinition(section_name: string; ident: string = 'layers'; are_output: boolean = false): TLayerArray;
  var
    a: TStringArray;
    s: string;
    i, id: integer;
    o: TAbstractLayer;
  begin
    SetLength(Result, 1);
    Result[0]:= nil;

    if (ident = 'layers') and are_output then ident:= 'output';

    s:= Trim(ReadString(section_name, ident, ''));
    if s = '' then
      if ident = 'layers'
      then Die('[' + section_name + ']:' + ident + ' not specified!')
      else if (ident = 'output') and are_output then begin
        OutStr('[' + section_name + ']:' + ident + ' not specified, assuming one default output layer named "sky".');
        s:= 'sky';
      end
      else Exit;

    a:= Explode(s, are_output);
    if Length(a) = 0 then
      if (ident = 'layers') or (ident = 'output')
        then Die('[' + section_name + ']:' + ident + ' has no layer names listed!')
        else Exit;

    if not are_output then begin
      SetLength(Result, Length(a) + 1);
      for i:= 0 to High(a) do begin
        if a[i] = 'project' then Die('Naming layers "project" is not allowed! ([' + section_name + ']:' + ident + ')');

        id:= List.IndexOf(a[i]);
        if id < 0 then begin
          o:= LayerFactory(a[i]);
          List.AddObject(a[i], o);
        end
        else
          o:= List.Objects[id] as TAbstractLayer;
        Result[i]:= o;
      end;
      Result[High(Result)]:= nil;  // because we treat it as a
      // nil-terminateed pointer chain when passing from layer to layer
    end
    else begin
      Result:= nil;
      for i:= 0 to High(a) do begin
        if OReserved.IndexOf(LowerCase(a[i])) < 0 then OReserved.Add(LowerCase(a[i]));
        if List.IndexOf(a[i]) >= 0 then Die('Naming collision: attempting to create output layer "' + a[i] + '" while an input layer with this name already exists!');
        OutputLayerFactory(a[i]);
      end;
    end;
  end;


  function TSkymp.ParseColorDefinition(section_name: string; ident: string = 'color'): TColorFPRec;
  var
     a: TStringArray;
     s: string;
     i: integer;
     f: float;
  begin
     s:= ReadString(section_name, ident, '');
     if s = '' then Die('[' + section_name + ']:' + ident + ' not specified!');
     a:= Explode(s);
     if (Length(a) < 3) or (Length(a) > 4)
       then Die('[' + section_name + ']:' + ident + ' invalid color definition, must have 3 or 4 floating point numbers separated by commas!');
     Result.a:= 1;
     for i:= 0 to High(a) do begin
       try
         f:= StrToFloat(a[i]);
       except
         if i < 3
           then  Die('[' + section_name + ']:' + ident + ' invalid color definition, "' + a[i] + '" is not a valid floating point number!')
           else continue;
       end;
       case i of
         0: Result.r:= f;
         1: Result.g:= f;
         2: Result.b:= f;
         3: Result.a:= f;
       end;
     end;
  end;

  function TSkymp.LayerFactory(name: string): TAbstractLayer;
  var
    c: CAbstractLayer;
    s: string;
  begin
    OutStr('Constructing layer ' + name + ' ...');

    s:= LowerCase(Trim(ReadString(name, 'type', '')));
    if s = '' then Die('[' + name + ']:type is not specified!');
    case s of
      'group': c:= TLayerGroup;
      'sphere_map': c:= TSphereMapLayer;
      'sprite': c:= TSpriteLayer;
      'color': c:= TSolidColorLayer;
      'gradient': c:= TGradientLayer;
    else
      Die('[' + name + ']:type has unsupported value of ' + s + ' ! Valid values are: group sphere_map sprite color gradient');
    end;

    Result:= c.Create(name, Self);

    OutStr('..finished constructing ' + name + ' ...');
  end;

  function IsDescendantOf(c1, c2: TClass): boolean;
  begin
    Result:= c1 = c2;
    while Assigned(c1) and (c1 <> c2) do begin
      c1:= c1.ClassParent;
      if c1 = c2 then begin
        Result:= true;
        break;
      end;
    end;
  end;

  procedure TSkymp.OutputLayerFactory(name: string);
  var
    c: CAbstractOutputLayer;
    cc: COutCubeMapLayer;
    s: string;
    m, i: integer;
    is_cubemap: boolean;
  begin
    OutStr('Constructing output ' + name + ' ...');
    is_cubemap:= false;

    s:= LowerCase(Trim(ReadString(LowerCase(name), 'type', '')));
    if s = '' then begin
      OutStr('  [' + name + ']:type is not specified, assuming sphere_map');
      s:= 'sphere_map';
    end;

    case s of
      'gzdoom_sky': c:= TOutDoomSkyLayer;
      'sphere_map': c:= TOutSphereMapLayer;
      'q3_cube_map': begin
        cc:= TOutCubeMapLayer;
        is_cubemap:= true;
      end
    else
      Die('[' + name + ']:type has unsupported value of ' + s
        + ' ! Valid values are: sphere_map q3_cube_map gzdoom_sky');
    end;

    m:= Length(OutLayers);
    //if c is TOutCubeMapLayer then begin
    //Frak. Error: Class or COM interface type expected, but got "CAbstractOutputLayer"
    if is_cubemap then begin//IsDescendantOf(c, TOutCubeMapLayer) then begin
      SetLength(OutLayers, Length(OutLayers) + 6);
      for i:= 0 to 5 do OutLayers[m + i]:= cc.Create(name, self, i);
    end
    else begin
      SetLength(OutLayers, Length(OutLayers) + 1);
      OutLayers[m]:= c.Create(name, Self);
    end;
  end;

  procedure TSkymp.OutStr(s: string);
  begin
    WriteLn(s);
  end;

  procedure TSkymp.ProgressCallback(p: float);
  var
    curprogress: integer;
  begin
    curprogress:= round(p * 50);
    if curprogress = prevprogress then Exit;
    Write('.');
    prevprogress:= curprogress;
  end;

  function TSkymp.IsRelativePath(path: string): boolean;
  begin
    if Length(path) < 1
      then Result:= true
      else Result:= path[1] = '.';
  end;

  constructor TSkymp.Create(fname: string);
  var i: integer;
  begin
    OutStr('Loading sky definition ' + fname) ;
    inherited Create(fname);

    is_debug:= ReadBool('project', 'debug', false);
    if (is_debug) then OutStr('DEBUGGING MODE: any exceptions during bitmap generation would be screened and shown as magenta pixels!');

    DefaultFormatSettings.DecimalSeparator:= '.';

    List:= TStringList.Create;
    List.OwnsObjects:= true;

    OReserved:= TStringList.Create;

    inpath:= ReadString('project', 'input_path', '');
    if (inpath = '') then begin
      OutStr('[project]:input_path is not specified!');
      inpath:= ExtractFilePath(fname);
    end
    else begin
      if IsRelativePath(inpath) then begin
        OutStr('[project]:input_path is declared as relative');
        inpath:= IncludeTrailingPathDelimiter(ExtractFilePath(fname)) + inpath;
      end;
    end;
    inpath:= IncludeTrailingPathDelimiter(inpath);
    OutStr('Input path is ' + inpath);

    outpath:= ReadString('project', 'output_path', '');
    if (outpath = '') then begin
      OutStr('[project]:output_path is not specified!');
      outpath:= ExtractFilePath(fname);
    end
    else begin
      if IsRelativePath(outpath) then begin
        OutStr('[project]:input_path is declared as relative');
        outpath:= IncludeTrailingPathDelimiter(ExtractFilePath(fname)) + outpath;
      end;
    end;
    outpath:= IncludeTrailingPathDelimiter(outpath);
    OutStr('Output path is ' + outpath);

    outname:= ChangeFileExt(ExtractFileName(fname), '');

    outext:= ReadString('project', 'output_image_format', '');
    i:= Pos('.', outext);
    if i > 0 then outext:= COPY(outext, i + 1, Length(outext) - i);
    outext:= Trim(LowerCase(outext));
    if outext = '' then begin
      OutStr('[project]:output_image_format is not specified, assuming tga');
      outext:= 'tga';
    end;

    supersamples:= max(0, ReadInteger('project', 'supersampling', 0));
    if supersamples = 0
    then OutStr('No supersampling')
    else if supersamples < 5 then begin
      OutStr('[project]:supersampling has invalid value ' + IntToStr(supersamples) + ', forcing 5');
      supersamples:= 5;
    end
    else if supersamples > 100 then begin
      OutStr('[project]:supersampling has invalid value ' + IntToStr(supersamples) + ', clamping to 100');
      supersamples:= 100;
    end
    else OutStr('Supersampling ' + IntToStr(supersamples));

    SetLength(OutLayers, 0);
    OutStr('Creating output layers');
    ParseLayerDefinition('project', 'output', true);
  end;

  function TSkymp.Execute: boolean;
  var i: integer;
  begin
    for i:= 0 to High(OutLayers) do begin
      OutStr('Generating output bitmap ' + IntToStr(i + 1) + '/' + IntToStr(Length(OutLayers)) + ' ...' );
      (OutLayers[i] as TAbstractOutputLayer).Iterate;
      TAbstractOutputLayer(OutLayers[i]).SaveBitmap;
    end;
    Result:= true;
  end;

  procedure TSkymp.Die(errormessage: string);
  begin
    raise ESkympDie.Create(errormessage);
  end;

  function Explode(s: string; NoLowerCase: boolean = false): TStringArray;
  var
    i, a: integer;
    t: string;
  begin
    Result:= nil;
    if s = '' then exit;
    if not NoLowerCase then s:= LowerCase(s);
    i:= 1;
    a:= 1;
    repeat
      if (s[i] = ',') then begin
        t:= Trim(COPY(s, a, i - a));
        if t <> '' then begin
          SetLength(Result, Length(Result) + 1);
          Result[High(Result)]:= t;
        end;
        i+= 1;
        a:= i;
      end
      else i+= 1;
    until i > length(s);
    if (a <= length(s)) then begin
      t:= Trim(COPY(s, a, length(s) - a + 1));
      if t <> '' then begin
        SetLength(Result, Length(Result) + 1);
        Result[High(Result)]:= t;
      end;
    end
  end;


  function LanczosWeight(Value: Single): Single;
    function SinC(Value: Single): Single;
    begin
      if Value <> float(0.0) then
      begin
        Value := Value * float(3.1415926535);//Pi;
        Result := Sin(Value) / Value;
      end
      else
        Result := float(1.0);
    end;
  begin
    if Value < float(0.0) then
      Value := -Value;
    if Value < float(3.0) then
      Result := SinC(Value) * SinC(Value / float(3.0))
    else
      Result := float(0.0);
  end;

  procedure TAbstractBitmapLayer.GetInterpolatedPixel(const x, y: float;  out color: TColorFPRec);
  var
    acc, col: TColorFPRec;
    xi, yi, xc, yc, bx, by: integer;
    weight, accWeight, fx, fy, rf: float;
  begin
    if x < 0 then begin
      xi:= trunc(x) - 1;
      fx:= 1.0 + frac(x);
    end
    else begin
      xi:= trunc(x);
      fx:= frac(x);
    end;
    if y < 0 then begin
      yi:= trunc(y) - 1;
      fy:= 1.0 + frac(y);
    end
    else begin
      yi:= trunc(y);
      fy:= frac(y);
    end;
    FillChar(acc, sizeof(acc), 0);
    accWeight:= 0;
    for xc:= -2 to 3 do
      for yc:= -2 to 3 do begin
        weight:= LanczosWeight((xc * 1.0) - fx);
        if weight = 0.0 then continue;
        weight*= LanczosWeight((yc * 1.0) - fy);
        if weight = 0.0 then continue;
        accWeight+= weight;
        if not WrapBitmapCoordinates(xc + xi, yc + yi, bx, by) then continue;
        col:= GetPixelFP(image, bx, by);
        acc.b+= col.b * weight;
        acc.g+= col.g * weight;
        acc.r+= col.r * weight;
        acc.a+= col.a * weight;
      end;
    if accWeight > 0.0001 then begin
      rf:= 1.0 / accWeight;
      color.b:= acc.b * rf;
      color.g:= acc.g * rf;
      color.r:= acc.r * rf;
      color.a:= acc.a * rf;
    end
    else FillChar(color, sizeof(color), 0);
  end;

  procedure TAbstractOutputLayer.Iterate;
  var
    x, y: integer;
    vector: TVector3f;
    coneWidth: float;
    color: TColorFPRec;
    first,next: PLayer;
  begin
    skymp.ProgressCallback(0.0);
    first:= @layers[0];
    for y:= 0 to image.height - 1 do begin
      for x:=0 to image.width - 1 do begin
        if skymp.is_debug then begin
          try
            BitmapToProjectVector(x, y, vector);
            coneWidth:= VectorToPixelConeWidth(vector);
            if skymp.supersamples > 0
              then SampleCone(first, vector, coneWidth, coneWidth, color)
              else begin
                next:= first;
                inc(next);
                first^.Sample(next, vector, color);
              end;
          except
            color.r:= 1.0;
            color.g:= 0.0;
            color.b:= 1.0;
            color.a:= 1.0;
          end;
        end
        else begin
          BitmapToProjectVector(x, y, vector);
          coneWidth:= VectorToPixelConeWidth(vector);
          if skymp.supersamples > 0
            then SampleCone(first, vector, coneWidth, coneWidth, color)
            else begin
              next:= first;
              inc(next);
              first^.Sample(next, vector, color);
            end;
        end;
        if flip
          then SetPixelFP(image, image.width - 1 - x, y, color)
          else SetPixelFP(image, x, y, color);
      end;
      skymp.ProgressCallback((y + 1) / image.height);
    end;
  end;

  constructor TAbstractOutputLayer.Create(_name: string; _skymp: TSkymp);
  var
    i: integer;
  begin
    inherited Create(LowerCase(_name), _skymp);

    inpath:= skymp.ReadString(name, 'input_path', '');
    if (inpath = '') then inpath:= skymp.inpath
    else begin
      if skymp.IsRelativePath(inpath) then begin
        skymp.OutStr('  [' + name + ']:input_path is declared as relative');
        inpath:= IncludeTrailingPathDelimiter(ExtractFilePath(skymp.FileName)) + inpath;
      end;
      inpath:= IncludeTrailingPathDelimiter(inpath);
      skymp.OutStr('  Input path redeclared as ' + inpath);
    end;

    outpath:= skymp.ReadString(name, 'output_path', '');
    if (outpath = '') then outpath:= skymp.outpath
    else begin
      if skymp.IsRelativePath(outpath) then begin
        skymp.OutStr('  [' + name + ']:input_path is declared as relative');
        outpath:= IncludeTrailingPathDelimiter(ExtractFilePath(skymp.FileName)) + outpath;
      end;
      outpath:= IncludeTrailingPathDelimiter(outpath);
      skymp.OutStr('  Output path redeclared as ' + outpath);
    end;

    outname:= name;

    outext:= skymp.ReadString(name, 'output_image_format', '');
    i:= Pos('.', outext);
    if i > 0 then outext:= COPY(outext, i + 1, Length(outext) - i);
    outext:= Trim(LowerCase(outext));
    if outext = '' then outext:= skymp.outext
    else begin
      skymp.OutStr('  Output image format redeclared as ' + outext);
    end;

    supersamples:= max(0, skymp.ReadInteger(name, 'supersampling', 0));
    if supersamples = 0
    then supersamples:= skymp.supersamples
    else if supersamples < 5 then begin
      skymp.OutStr('  [' + name + ']:supersampling has invalid value ' + IntToStr(supersamples) + ', forcing 5');
      supersamples:= 5;
    end
    else if supersamples > 100 then begin
      skymp.OutStr('  [' + name + ']:supersampling has invalid value ' + IntToStr(supersamples) + ', clamping to 100');
      supersamples:= 100;
    end
    else skymp.OutStr('  Supersampling ' + IntToStr(supersamples));

    Layers:= skymp.ParseLayerDefinition(name);

    filename:= outpath + outname + '.' + outext;
  end;


  constructor TOutSphereMapLayer.Create(_name: string; _skymp: TSkymp);
  begin
    inherited Create(_name, _skymp);
    equwidth:= image.width;
    equheight:= image.height;
  end;

  constructor TOutCubeMapLayer.Create(_name: string; _skymp: TSkymp; _num: integer);
  var
    suff: string;
    m1, m2: TMatrix4f;
    v: Tvector3f;
  begin
//    skymp.OutStr('_num=' + IntToStr(num));
    inherited Create(_name, _skymp);
    num:= _num;
    skymp.equwidth:= image.width * 4;
    skymp.equheight:= image.height * 2;
    LoadIdentity(matrix);
    suff:= '';
    case _num of
      0: begin
          suff:= '_ft';
          LoadIdentity(matrix);
        end;
      1: begin
          suff:= '_lf';
          FillChar(v, sizeof(v), 0);
          v[1]:= -1;
          Rotate(matrix, Pi / 2, v);
        end;
      2: begin
          suff:= '_rt';
          FillChar(v, sizeof(v), 0);
          v[1]:= -1;
          Rotate(matrix, - Pi / 2, v);
        end;
      3: begin
          suff:= '_bk';
          FillChar(v, sizeof(v), 0);
          v[1]:= -1;
          Rotate(matrix, Pi, v);
        end;
      4: begin
          suff:= '_up';
          FillChar(v, sizeof(v), 0);
          v[0]:= 1;
          Rotate(m2, - Pi/2, v);
          m1:= matrix;
          matrix:= m1 * m2;

          FillChar(v, sizeof(v), 0);
          v[2]:= -1;
          Rotate(m2, Pi/2, v);
          m1:= matrix;
          matrix:= m1 * m2;
        end;
      5: begin
          suff:= '_dn';
          FillChar(v, sizeof(v), 0);
          v[0]:= 1;
          Rotate(m2, Pi/2, v);
          m1:= matrix;
          matrix:= m1 * m2;

          FillChar(v, sizeof(v), 0);
          v[2]:= -1;
          Rotate(m2, -Pi/2, v);
          m1:= matrix;
          matrix:= m1 * m2;
       end;
    end;
    filename:= outpath + outname + suff + '.' + outext;
//skymp.OutStr('filename=' + filename);
  end;

  procedure TOutCubeMapLayer.BitmapToProjectVector(const x, y: float; out vector: TVector3f);
  begin
    FillChar(vector, sizeof(vector), 0);
    vector[0]:= sin((Pi / 2) * (0.5 - (x / (image.width - 1))));
    vector[1]:= sin((Pi / 2) * (0.5 - (y / (image.height - 1))));
    vector[2]:= - 1 / sqrt(2);
    Normalize(vector);
    vector*= matrix;
  end;







  function TOutCubeMapLayer.VectorToPixelConeWidth(const vector: TVector3f): float;
  var v1, v2: TVector3f;
  begin
    FillChar(v2, sizeof(v2), 0);
    v2*= matrix;
    v2[1]:= -1;
    v1:= vector;
    Result:= (sqrt(2) * Pi / (2 * image.width)) * DotProduct(v1, v2);
  end;



  procedure TAbstractlayer.SampleCone(next: PLayer; const vector: TVector3f; const coneWidth, pixelWidth: float; out color: TColorFPRec);
  var
    next2: PLayer;
    perv, rotv: TVector3f;
    pstep, ystep: float;
    iterations, i: integer;
  begin
    //STUB!
    next2:= next;
    inc(next2);
    next^.Sample(next2, vector, color);

    //FindHorizontalPerpendicularVector(vector, perv);

    // iterations:= skymp.supersamples;


    //here we iterate in a spiral pattern
    //first we take horizontal perpendicular vector, rotate it around our vector,
    //then in turn rotate our vector around it!
    //allows for insanlely wide cones - like, 270 degrees easily - usable,
    // among other things, for gaussian blur
  end;


  procedure LoadIdentity(var matrix: TMatrix4f);
  begin
    FillChar(matrix, sizeof(matrix), 0);
    matrix[0,0]:= 1;
    matrix[1,1]:= 1;
    matrix[2,2]:= 1;
    matrix[3,3]:= 1;
  end;


  procedure Rotate(var matrix: TMatrix4f; angle: float; const vector: TVector3f);
  //https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/glRotate.xml
  var s, c, x, y, z: float;
  begin
    s:= sin(angle);
    c:= cos(angle);
    x:= vector[0];
    y:= vector[1];
    z:= vector[2];
    matrix[0, 0]:= x * x * (1 - c) + c;
    matrix[0, 1]:= y * x * (1 - c) + z * s;
    matrix[0, 2]:= z * x * (1 - c) - y * s;
    matrix[0, 3]:= 0;

    matrix[1, 0]:= x * y * (1 - c) - z * s;
    matrix[1, 1]:= y * y * (1 - c) + c;
    matrix[1, 2]:= y * z * (1 - c) + x * s;
    matrix[1, 3]:= 0;

    matrix[2, 0]:= x * z * (1 - c) + y * s;
    matrix[2, 1]:= y * z * (1 - c) - x * s;
    matrix[2, 2]:= z * z * (1 - c) + c;
    matrix[2, 3]:= 0;

    matrix[3, 0]:= 0;
    matrix[3, 1]:= 0;
    matrix[3, 2]:= 0;
    matrix[3, 3]:= 1;
  end;

  procedure Ortho(var matrix: TMatrix4f; left, right, bottom, top,  nearVal, farVal: float);
  begin
    //https://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml
    matrix[0, 0]:= 2 / (right - left);
    matrix[0, 1]:= 0;
    matrix[0, 2]:= 0;
    matrix[0, 3]:= 0;

    matrix[1, 0]:= 0;
    matrix[1, 1]:= 2 / (top - bottom);
    matrix[1, 2]:= 0;
    matrix[1, 3]:= 0;

    matrix[2, 0]:= 0;
    matrix[2, 1]:= 1;
    matrix[2, 2]:= -2 / (farVal - nearVal);
    matrix[2, 3]:= 0;

    matrix[3, 0]:= - (right + left) / (right - left);
    matrix[3, 1]:= - (top + bottom) / (top - bottom);
    matrix[3, 2]:= - (farVal + nearVal) / (farVal - nearVal);
    matrix[3, 3]:= 1;
  end;

  procedure FindHorizontalPerpendicularVector(const v1: TVector3f; out v2: TVector3f);
  var f: float;
  begin
    f:= sqr(v1[0]) + sqr(v1[2]);
    if (f < float(0.0001)) then begin
      v2[0]:= 1;
      v2[1]:= 0;
      v2[2]:= 0;
      Exit;
    end;
    f:= 1.0 / sqrt(f);
    v2[0]:= v1[2] * f;
    v2[1]:= 0;
    v2[2]*= v1[0] * f;
  end;


  function TAbstractBitmapLayer.Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean;
  var
    v2: TVector3f;
    MyColor: TColorFPRec;
    bx, by: float;
  begin
    v2:= vector * matrix;
    if UnprojectVectorToBitmap(v2, bx, by) then begin
      //MyColor.a:= 1;
      //MyColor.b:= dbgb;
      //MyColor.R:= dbgr;//bx/image.width;
      //MyColor.G:= dbgg; //by/image.height;
      GetInterpolatedPixel(bx, by, MyColor);
    end
    else
      MyColor.a:= 0;
    Result:= SampleNextAndBlend(MyColor, next, vector, Color);
  end;


  function TLayerGroup.Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean;
  var
    MyColor: TColorFPRec;
    myfirst, mynext: PLayer;
  begin
    myfirst:= PLayer(@Layers[0]);
    mynext:= myfirst;
    inc(mynext);
    myfirst^.Sample(mynext, vector, MyColor);
    MyColor.a*= opacity;
    Result:= SampleNextAndBlend(MyColor, next, vector, Color);
  end;

  function TSolidColorLayer.Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean;
  begin
    Result:= SampleNextAndBlend(MyColor, next, vector, Color);
  end;


  function TAbstractLayer.SampleNextAndBlend(const MyColor: TColorFPRec; next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean;
  var
    Next2: PLayer;
    NextColor: TColorFPRec;
  begin
    if MyColor.a > 0.0001 then begin
       // has own color to mix
       Result:= true;
       if Assigned(next) and Assigned(next^) then begin
         next2:= next;
         inc(next2);
         next^.Sample(next2, vector, NextColor);
       end
       else begin
         Result:= false;
         FillChar(NextColor, sizeof(NextColor), 0);
       end;
       Color:= Blend(MyColor, NextColor);
     end
     else begin
       // the vector projection is out of bitmap bounds. Passthru from the next layer.
       if Assigned(next) and Assigned(next^) then begin
         next2:= next;
         inc(next2);
         Result:= next^.Sample(next2, vector, Color);
       end
       else begin
         Result:= false;
         FillChar(Color, sizeof(Color), 0);
       end;
     end;
  end;

  function TSphereMapLayer.UnprojectVectorToBitmap(const vector: TVector3f; out x, y: float): boolean;
  var
    v2: TVector3f;
    pitch, yaw, f: float;
  begin
    Result:= true; // always hits
    v2:= vector;// * revmat;
    Normalize(v2); // y points straight up

    pitch:= arcsin(vector[1]);
    y:= (0.5 - pitch / pi) * (image.height - 1);
    f:= sqr(v2[0]) + sqr(v2[2]);
    if (f < 0.0000001) then begin
      x:= 0;
      Exit;
    end;
    f:= 1.0 / sqrt(f);
    if vector[2] < 0
      then yaw:= arcsin(f * vector[0])
      else yaw:= Pi - arcsin(f * vector[0]);
    if flip then begin
       if NonClosedHorizontally
         then x:= (0.5 - yaw / (2 * pi)) * (image.width)
         else x:= (0.5 - yaw / (2 * pi)) * (image.width - 1);
    end
    else begin
      if NonClosedHorizontally
        then x:= (0.5 + yaw / (2 * pi)) * (image.width)
        else x:= (0.5 + yaw / (2 * pi)) * (image.width - 1);
    end;
  end;

  procedure TSphereMapLayer.BitmapToProjectVector(const x, y: float; out vector: TVector3f);
  var pitch, yaw, ax, az: float;
  begin
    //vector is in opengl coordinates (x right, y up, z pokes at your eye)
    //bitmap coordinates are 0,0 = top left corner
    if NonClosedHorizontally
      then yaw:= (-0.5 + (x / image.width)) * 2 * Pi
      else yaw:= (-0.5 + (x / (image.width - 1))) * 2 * Pi;
    pitch:= (0.5 - (y / (image.height - 1))) * Pi;
    vector[0]:= cos(pitch) * sin(yaw);
    vector[1]:= sin(pitch);
    vector[2]:= cos(pitch) * - cos(yaw);
  end;


  procedure TOutSphereMapLayer.BitmapToProjectVector(const x, y: float; out vector: TVector3f);
  var pitch, yaw, ax, az: float;
  begin
    //vector is in opengl coordinates (x right, y up, z pokes at your eye)
    //bitmap coordinates are 0,0 = top left corner
    if NonClosedHorizontally
      then yaw:= (-0.5 + (x / image.width)) * 2 * Pi
      else yaw:= (-0.5 + (x / (image.width - 1))) * 2 * Pi;
    pitch:= (0.5 - (y / (image.height - 1))) * Pi;
    vector[0]:= cos(pitch) * sin(yaw);
    vector[1]:= sin(pitch);
    vector[2]:= cos(pitch) * - cos(yaw);
  end;

  const
    GZDoomSkyHeight = 512;
    GZDoomSkyMiddle = 207.5;
    GZDoomSkyTopPart = (90 - 22) / 90; // 22 degrees lost!
    GZDoomSkyBottomPart = (90 - 15) / 90;

  procedure TOutDoomSkyLayer.BitmapToProjectVector(const x, y: float; out vector: TVector3f);
  var
    pitch, ynl, nlf, yaw, ax, az: float;
  begin
    yaw:= (0 - (x / (image.width - 1))) * 2 * Pi;
    // this shit is *grossly* non-linear!!!
    // top part of sky is half of 415 pixels, with 20 degrees in zenith lost,
    //  the game engine fills it with blurred color derived from the top image edge.
    if y < GZDoomSkyMiddle  then begin
        ynl:= (GZDoomSkyMiddle - y) * (GZDoomSkyTopPart / (GZDoomSkyMiddle * 2));
        //it is stretched non-linearly! The middle is larger.
        nlf:= 1 + (0.5 - ynl / GZDoomSkyTopPart)  * 0.5 * 2;
        ynl*= nlf;
    end
    // nadir. The same story here, blurred color derived from the bottom image edge
    else begin
      ynl := - (y - GZDoomSkyMiddle) * (GZDoomSkyBottomPart / ((GZDoomSkyHeight - 1 - GZDoomSkyMiddle) * 2));
      nlf:= 1 + (0.5 + ynl / GZDoomSkyBottomPart)  * 0.78 * 2;
      ynl*= nlf;
    end;
    pitch:= ynl * Pi;
    vector[0]:= cos(pitch) * sin(yaw);
    vector[1]:= sin(pitch);
    vector[2]:= cos(pitch) * - cos(yaw);
  end;

  procedure TOutDoomSkyLayer.ParseDimensions(out w, h: integer);
  begin
    Skymp.OutStr('  GZDoom sky: size 1024x' + IntToStr(GZDoomSkyHeight) + ', zenith and nadir not covered!');
    w:= 1024;
    h:= GZDoomSkyHeight;
    NonClosedHorizontally:= true;
  end;

  function TSphereMapLayer.VectorToPixelConeWidth(const vector: TVector3f): float;
  begin
    //independent of the view direction. Horizontal stretching is not taken into account
    Result:= min(1 / image.height, 1 / image.width);
  end;

  function TOutSphereMapLayer.VectorToPixelConeWidth(const vector: TVector3f): float;
  begin
    Result:= min(1 / image.height, 1 / image.width);
  end;

  function TSphereMapLayer.WrapBitmapCoordinates(x, y: integer; out x1, y1: integer): boolean;
  var
    a, b, wn, iw: integer;
  begin
    Result:= true; //always hits
    if NonClosedHorizontally
      then iw:= image.width
      else iw:= image.width - 1;
    if y < 0 then begin
      wn := (-y) div image.height;
      b:= (-y) - wn * image.height;
      if (wn and 1) = 1 then begin
        //each second wrap around it passes the opposite part of the map in opposite direction
        // at y = 0 (zenith) and y = image.height - 1 (nadir) the x coordinate is,
        //  strictly speaking, meaningless and all pixels of the top and bottom image
        //  edges must have the same color. This is not checked, though.
        x+= iw div 2;
        y1:= b - 1;
      end
      else
        y1:= image.height - b;
    end else begin
      wn:= y div image.height;
      if (wn and 1) = 1 then begin
        x+= iw div 2;
        y1:= image.height - (y mod image.height) - 1;
      end
      else
        y1:= y mod image.height;
    end;
    if (x < 0) then begin
      x1:= iw - ((-x) mod iw) - 1;
    end
    else
      x1:= x mod iw;
  end;


  constructor TAbstractLayer.Create(_name: string; _skymp: TSkymp);
  var
    i: integer;
  begin
    inherited Create;
    skymp:= _skymp;
    fname:= _name;
    mode:= skymp.ParseMode(_name);

  end;

  procedure TAbstractLayer.ParseOpacityDefinition;
  var
    str: string;
  begin
    str:= skymp.ReadString(name, 'opacity', '');
    if str = '' then begin
      skymp.OutStr('  [' + fname + ']:opacity is not specified, assuming 1.0');
      opacity:= 1;
    end
    else begin
      opacity:= skymp.ReadFloat(name, 'opacity', 1.0);
      skymp.OutStr('  opacity = ' + FloatToStrf(opacity, ffGeneral, 1,4));
    end;
  end;


  function signtostr(f: float; pos, neg: string): string;
  begin
    if f >= 0.0 then Result:= pos else Result:= neg;
  end;


  constructor TAbstractBitmapLayer.Create(_name: string; _skymp: TSkymp);
  var
    m1, m2: TMatrix4f;
    rv: TVector3f;
    iw, s, roll, pitch, yaw: float;
    str: string;
  begin
    inherited Create(_name, _skymp);
    LoadIdentity(matrix);
    ProduceImage;

    if (_name = 'sky') or (_name = '')
      then opacity:= 1
      else ParseOpacityDefinition;

    flip:= skymp.readBool(name, 'flip', false);
    if flip then skymp.OutStr('  flipped.');

    roll:= Pi / 180 * skymp.ReadFloat(name, 'roll', 0.0);
    if roll <> 0.0 then begin
      skymp.OutStr('  roll = ' + FloatToStrf(abs(roll) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(roll, 'clockwise', 'counterclockwise'));
      FillChar(rv, sizeof(rv), 0);
      rv[2]:= -1;
      Rotate(m2, roll, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;



    iw:= image.width;

    pitch:= (Pi / 180) * skymp.ReadFloat(name, 'pitch', 0.0);
    if pitch = 0.0 then begin
      s:= skymp.ReadFloat(name, 'y_shift', 0.0);
      pitch:= (Pi / (image.height - 1)) * s;
      if pitch <> 0.0 then skymp.OutStr('  pitch is defined from [' + fname + ']:y_shift = ' + FloatToStrf(s ,ffGeneral,4,2) + ' pixels');
    end;
    if pitch <> 0.0 then begin
      skymp.OutStr('  pitch = ' + FloatToStrf(abs(pitch) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(pitch, 'upwards', 'downwards'));
      FillChar(rv, sizeof(rv), 0);
      rv[0]:= 1;
      Rotate(m2, pitch, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;




    yaw:= (Pi / 180) * skymp.ReadFloat(name, 'yaw', 0.0);
    if yaw = 0.0 then begin
      s:= skymp.ReadFloat(name, 'x_shift', 0.0);
      yaw:= (Pi / iw) * s;
      if yaw <> 0.0 then skymp.OutStr('  yaw is defined from [' + fname + ']:x_shift = ' + FloatToStrf(s ,ffGeneral,4,2) + ' pixels');
    end;
    if yaw <> 0.0 then begin
      skymp.OutStr('  yaw = ' + FloatToStrf(abs(yaw) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(yaw, 'rightwise', 'leftwise'));
      FillChar(rv, sizeof(rv), 0);
      rv[1]:= -1;
      Rotate(m2, yaw, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;



//    if NonClosedHorizontally
//      then iw:= image.width
//      else iw:= image.width - 1;



    Invert(matrix, revmat);
  end;


  procedure TSpriteLayer.BitmapToProjectVector(const x, y: float; out vector: TVector3f);
  begin
    skymp.Die('Method BitmapToProjectVector() should never be called on TSpriteLayer class!');
  end;

  function TSpriteLayer.VectorToPixelConeWidth(const vector: TVector3f): float;
  begin
    Result:= min(xangle / image.width, yangle / image.height);
  end;

  function TSpriteLayer.WrapBitmapCoordinates(x, y: integer; out x1, y1: integer): boolean;
  begin
    if (x < 0) or (x > image.width - 1) or (y < 0) or (y > image.height - 1)
      then Exit(false);
    x1:= x;
    y1:= y;
    Result:= true;
  end;

  const
    ForwardVector: TVector3f = (0, 0, -1);
    UpwardVector: TVector3f = (0, 1, 0);



  function TSpriteLayer.UnprojectVectorToBitmap(const vector: TVector3f; out x, y: float): boolean;
  var
    v2, forward: TVector3f;
    aa, sar, ar, ah, ax, ay, aac: float;
  begin
    v2:= vector;
    Normalize(v2);
    // aa is the angle between forward axis and the vector
    aa:= arccos(DotProduct(v2, ForwardVector));
//dbgr:= 0; //0.5 + aa * 0.5;
    if (sqr(v2[0]) + sqr(v2[1])) > float(0.00001) then begin
      v2[2]:= 0;
      Normalize(v2);
      // ar = 0 points upwards, clockwise
      ar:= arccos(v2[1]) * sign(v2[0]);
//dbgb:= 0; //0.5 + ar*0.5;
      //ax:= aa * sin(ar);
      //ay:= aa * cos(ar); // such NASTY instability at downward angles!
      sar:= sin(ar);
      if sar > float(0.00001) then begin
        ax:= aa * sar;
        ay:= aa * cos(ar);
      end
      else begin
        ax:= aa * sar;
        ah:= arccos(v2[0]) * sign(v2[1]);
        ay:= aa * sin(ah);
      end;
//dbgg:= 0.5 + ay*0.5;
      if (abs(ax) * 2 > xangle) or (abs(ay) * 2 > yangle) then begin
        Result:= false;
      end
      else begin
        if flip
          then x:= (0.5 - ax / xangle) * (image.width - 1)
          else x:= (0.5 + ax / xangle) * (image.width - 1);
        y:= (0.5 - ay / yangle) * (image.height - 1);
        Result:= true;
      end;
    end
    else begin
      if v2[2] < float(0) then begin
//dbgg:= 0.5;
//dbgb:= 0.5;
        x:= (image.width - 1) / 2;
        y:= (image.height - 1) / 2;
        Result:= true;
      end else begin
        Result:= false; // is antipodal dot.
      end;
    end;
  end;

  constructor TGradientLayer.Create(_name: string; _skymp: TSkymp);
  var
    s: string;
    m1, m2: TMatrix4f;
    roll, pitch, yaw: float;
    rv: TVector3f;
  begin
    inherited Create(_name, _skymp);

    skymp.OutStr('  Solid color.');
    ParseOpacityDefinition;
    Color1:= skymp.ParseColorDefinition(_name, 'start_color');
    Color1.a*= opacity;
    skymp.OutStr('  Start color: red = ' + FloatToStrf(Color1.r, ffGeneral, 1,4)
      + ', greeng = ' + FloatToStrf(Color1.g, ffGeneral,1,4)
      + ', blue = ' + FloatToStrf(Color1.b, ffGeneral,1,4)
      + ', alpha = ' + FloatToStrf(Color1.a, ffGeneral,1,4)
     );

    s:= trim(skymp.ReadString(name, 'end_color', ''));
    if s = '' then begin
      skymp.OutStr('  [' + name + ']:end_color not specified. Assuming the same as the start color but with zero alpha.');
      Color2:= Color1;
      Color2.a:= 0;
    end
    else begin
      Color2:= skymp.ParseColorDefinition(_name, 'end_color');
      Color2.a*= opacity;
      skymp.OutStr('  End color: red = ' + FloatToStrf(Color2.r, ffGeneral, 1,4)
        + ', greeng = ' + FloatToStrf(Color2.g, ffGeneral,1,4)
        + ', blue = ' + FloatToStrf(Color2.b, ffGeneral,1,4)
        + ', alpha = ' + FloatToStrf(Color2.a, ffGeneral,1,4)
       );
    end;

    pow:= skymp.ReadFloat(name, 'power', 0.0);
    if pow < 0.0001 then begin
      skymp.OutStr('  [' + name + ']:power not specified, assuming 1');
      pow:= 1;
    end
    else
      skymp.OutStr('  power = ' +  FloatToStrf(pow ,ffGeneral,3,3));

    angle1:= max(0.0, min(2 * Pi, (Pi / 180) * skymp.ReadFloat(name, 'start_angle', 0.0)));
    skymp.OutStr('  start angle = ' +  FloatToStrf(angle1 * 180 /Pi ,ffGeneral,3,2) + ' degrees');
    angle2:= max(0.0, min(2 * Pi, (Pi / 180) * skymp.ReadFloat(name, 'end_angle', 0.0)));
    skymp.OutStr('  end angle = ' +  FloatToStrf(angle2 * 180 /Pi ,ffGeneral,3,2) + ' degrees');
    if abs (angle1 - angle2) < 0.0001 then skymp.Die('[' + name + ']:start_angle and [' + name  + ']:end_angle must not be equal!');

    LoadIdentity(matrix);

    roll:= Pi / 180 * skymp.ReadFloat(name, 'roll', 0.0);
    if roll <> 0.0 then begin
      skymp.OutStr('  roll = ' + FloatToStrf(abs(roll) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(roll, 'clockwise', 'counterclockwise'));
      FillChar(rv, sizeof(rv), 0);
      rv[2]:= -1;
      Rotate(m2, roll, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;

    pitch:= (Pi / 180) * skymp.ReadFloat(name, 'pitch', 0.0);
    if pitch <> 0.0 then begin
      skymp.OutStr('  pitch = ' + FloatToStrf(abs(pitch) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(pitch, 'upwards', 'downwards'));
      FillChar(rv, sizeof(rv), 0);
      rv[0]:= 1;
      Rotate(m2, pitch, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;

    yaw:= (Pi / 180) * skymp.ReadFloat(name, 'yaw', 0.0);
    if yaw <> 0.0 then begin
      skymp.OutStr('  yaw = ' + FloatToStrf(abs(yaw) * 180 /Pi ,ffGeneral,3,2) + ' degrees ' + signtostr(yaw, 'rightwise', 'leftwise'));
      FillChar(rv, sizeof(rv), 0);
      rv[1]:= -1;
      Rotate(m2, yaw, rv);
      m1:= matrix;
      matrix:= m1 * m2;
    end;
  end;

  function TGradientLayer.Sample(next: PLayer; const vector: TVector3f; out Color: TColorFPRec): boolean;
  var
    c: TColorFPRec;
    v2: TVector3f;
    a, f: float;
  begin
    result:= true;
    v2:= vector * matrix;
    Normalize(v2);
    // a is the angle between forward axis and the vector
    a:= arccos(DotProduct(v2, ForwardVector));
    if angle1 < angle2 then begin
      if a < angle1 then f:= 1 else if a > angle2 then f:= 0 else f:= 1 - (a - angle1) / (angle2 - angle1);
    end
    else begin
      if a < angle2 then f:= 0 else if a > angle1 then f:= 1 else f:= (a - angle2) / (angle1 - angle2);
    end;
    f:= power(f, pow);
    c.r:= Color1.r * f + Color2.r * (1 - f);
    c.g:= Color1.g * f + Color2.g * (1 - f);
    c.b:= Color1.b * f + Color2.b * (1 - f);
    c.a:= Color1.a * f + Color2.a * (1 - f);
    Result:= SampleNextAndBlend(c, next, vector, Color);
  end;


  constructor TSpriteLayer.Create(_name: string; _skymp: TSkymp);
  var
    a, fa, a1, a2, ma, aspect, b, s: float;
  begin
    inherited Create(_name, _skymp);
    aspect:= image.height / image.width;
    fa:= skymp.ReadFloat(name, 'force_aspect', 0.0);
    if fa > 0.0001 then begin
      aspect:= fa;
      skymp.OutStr('  [' +  name + ']:force_aspect in effect, ' + FloatToStrF(fa,ffGeneral,3,4));
    end
    else begin
      fa:= skymp.ReadFloat(name, 'aspect_correction', 0.0);
      if fa > 0.0001 then begin
        aspect*= fa;
        skymp.OutStr('  [' +  name + ']:aspect_correction in effect, resulting aspect = ' + FloatToStrF(aspect,ffGeneral,3,4));
      end;
    end;
    b:= min(aspect, 1 / aspect); // for calculating maximum allowed angle (corners meet at 360)
    ma:= 2 * Pi / sqrt(1 + sqr(b));
    a:= (Pi / 180) * skymp.ReadFloat(name, 'angle', 0.0);
    if a <= 0.0 then begin
      skymp.OutStr('  [' +  name + ']:angle not specified, using size in pixels instead');
      s:= skymp.ReadFloat(name, 'size', 0.0);
      if s <= 0.0 then begin
        s:= max(image.width, image.height);
        skymp.OutStr('  [' +  name + ']:size not specified, assuming ' + FloatToStrf(s,ffGeneral,3,2) + ' pixels');
      end;
      a1:= Pi * s / skymp.equheight;
      a2:= 2 * Pi * s / skymp.equwidth;
      a:= min(a1, a2);
      skymp.OutStr('  The derived angle is ' + FloatToStrf(a * 180 /Pi ,ffGeneral,3,2) + ' degrees');
    end
    else
      skymp.OutStr('  angle = ' + FloatToStrf(a * 180 /Pi ,ffGeneral,3,2) + ' degrees');
    if a >= ma then skymp.Die('Angle is greater than maximum allowed angle for this sprite aspect, which is '  + FloatToStrf(ma * 180 /Pi ,ffGeneral,3,2) + ' degrees');

    if aspect > 1 then begin
      yangle:= a;
      xangle:= a / aspect;
    end
    else begin
      xangle:= a;
      yangle:= a * aspect;
    end;
  end;


  procedure TAbstractBitmapLayer.ProduceImage;
  var
    fileName, dir: string;
  begin
    filename:= skymp.ReadString(fname, 'bitmap', '');
    if filename = '' then skymp.Die('[' + fname + ']:bitmap is not specified!');
    dir:= ExtractFilepath(fileName);
    if skymp.IsRelativepath(dir) then fileName:= skymp.inpath + fileName;
    skymp.OutStr('  Loading bitmap ' + fileName);
    if not FileExists(fileName) then skymp.Die('File not found!');
    if not LoadImageFromFile(fileName, image) then Skymp.Die('Failed to load the image!');
    Skymp.OutStr('  width=' + IntToStr(image.width) + '; height=' + IntToStr(image.height) + '; format=' + VyFormatName(image.format));
  end;

  function VyFormatName(i: TImageFormat) : String;
  begin
    case i of
      ifUnknown: Result:= 'Unknown';
      ifDefault: Result:= 'Default';
      { Indexed formats using palette.}
      ifIndex8: Result:= 'Indexed';
      { Grayscale/Luminance formats.}
      ifGray8: Result:= 'Luminance8';
      ifA8Gray8: Result:= 'LuminanceAlpha8';
      ifGray16: Result:= 'Luminance16';
      ifGray32: Result:= 'Luminance32';
      ifGray64: Result:= 'Luminance64';
      ifA16Gray16: Result:= '';
      { ARGB formats.}
      ifX5R1G1B1: Result:= 'X5R1G1B1';
      ifR3G3B2: Result:= 'R3G3B2';
      ifR5G6B5: Result:= 'R5G6B5';
      ifA1R5G5B5: Result:= 'A1R5G5B5';
      ifA4R4G4B4: Result:= 'RGBA4';
      ifX1R5G5B5: Result:= 'X1R5G5B5';
      ifX4R4G4B4: Result:= 'XRGB4';
      ifR8G8B8: Result:= 'RGB8';
      ifA8R8G8B8: Result:= 'RGBA8';
      ifX8R8G8B8: Result:= 'XRGB8';
      ifR16G16B16: Result:= 'RGB16';
      ifA16R16G16B16: Result:= 'RGBA16';
      ifB16G16R16: Result:= 'RGB16';
      ifA16B16G16R16: Result:= 'RGBA16';
      { Floating point formats.}
      ifR32F: Result:= 'R32F';
      ifA32R32G32B32F: Result:= 'RGBA32F';
      ifA32B32G32R32F: Result:= 'RGBA32F';
      ifR16F: Result:= 'R16F';
      ifA16R16G16B16F: Result:= 'RGBA16F';
      ifA16B16G16R16F: Result:= 'RGBA16F';
      { Special formats.}
      ifDXT1: Result:= 'DXT1';
      ifDXT3: Result:= 'DXT3';
      ifDXT5: Result:= 'DXT5';
      ifBTC: Result:= 'BTC';
      ifATI1N: Result:= 'ATI1N';
      ifATI2N: Result:= 'ATI2N';
    else
      Result:= IntToStr(ord(i));
    end;
  end;


  procedure TOutSphereMapLayer.ParseDimensions(out w, h: integer);
  begin
    h:= max (0, min(4096, skymp.ReadInteger(name, 'height', 0)));
    w:= max (0, min(8192, skymp.ReadInteger(name, 'sphere_width', 0)));
    if (w = 0) and (h = 0)  then skymp.Die('Both [' + name + ']:height and [' + name + ']:sphere_width are not specified!');
    if w = 0 then begin
      w:= max(16, h * 2);
      Skymp.OutStr('  [' + name + ']:sphere_width is not specified!');
    end;
    if h = 0 then begin
      h:= max(16, w div 2);
      skymp.OutStr('  [' + name + ']:height is not specified!');
    end;
    NonClosedHorizontally:= skymp.ReadBool(name, 'non_closed_horizontally', false);
    if not NonClosedHorizontally then Skymp.OutStr('  Sphere map is closed horizontally (no gap between right and left sides, edge pixels are identic)');
  end;


  procedure TOutCubeMapLayer.ParseDimensions(out w, h: integer);
  begin
    h:= max (0, min(4096, skymp.ReadInteger(name, 'height', 0)));
    if h = 0  then skymp.Die('[' + name + ']:height is not specified!');
    h:= max (16, h);
    w:= h;

    if flip then skymp.Die('NOT IMPLEMENTED YET: cube maps do not support flipping!')
  end;


  procedure TAbstractOutputLayer.ProduceImage;
  var
    w, h: integer;
    format_name: string;
    format: TImageFormat;
    a: boolean;
  begin
    skymp.OutStr('  Creating output bitmap.');
    ParseDimensions(w, h);
    if skymp.ReadBool(name, 'hdr', false) then begin
      skymp.OutStr('  [' + name + ']:hdr is true.');
      format:= ifA32B32G32R32F;
      a:= true;
    end
    else begin
      if skymp.ReadBool(name, 'alpha', false) then begin
        skymp.OutStr('  [' + name + ']:alpha is true.');
        format:= ifA8R8G8B8;
        a:= true;
      end
      else begin
        format:= ifR8G8B8;
        a:= false;
      end;
    end;
    format_name:= VyFormatName(format);
      //GetEnumName(typeinfo(TImageFormat), ord(format));
      //doesn't work, such enums do not have valid RTTI

    Skymp.OutStr('  width=' + IntToStr(w) + '; height=' + IntToStr(h) + '; format=' + format_name);

    if not NewImage(w, h, format, image) then skymp.Die('Failed to create empty image!');

    LoadIdentity(matrix);
    LoadIdentity(revmat);
  end;

  procedure TAbstractOutputLayer.SaveBitmap;
  begin
    Skymp.OutStr('Writing ' + filename);
    if not SaveImageToFile(filename, image) then skymp.Die('Failed to save the image!');
    Skymp.OutStr('OK');
  end;


  constructor TLayerGroup.Create(_name: string; _skymp: TSkymp);
  begin
    inherited Create(_name, _skymp);
    skymp.OutStr('  Layer group.');
    ParseOpacityDefinition;
    Layers:= skymp.ParseLayerDefinition(_name);
  end;

  constructor TSolidColorLayer.Create(_name: string; _skymp: TSkymp);
  begin
    inherited Create(_name, _skymp);
    skymp.OutStr('  Solid color.');
    ParseOpacityDefinition;
    MyColor:= skymp.ParseColorDefinition(_name);
    MyColor.a*= opacity;
    skymp.OutStr('  Solid color: red = ' + FloatToStrf(MyColor.r, ffGeneral, 1,4)
      + ', greeng = ' + FloatToStrf(MyColor.g, ffGeneral,1,4)
      + ', blue = ' + FloatToStrf(MyColor.b, ffGeneral,1,4)
      + ', alpha = ' + FloatToStrf(MyColor.a, ffGeneral,1,4)
     );
  end;

  constructor TSphereMapLayer.Create(_name: string; _skymp: TSkymp);
  begin
    inherited Create(_name, _skymp);
  end;

  function sRGBToLinear(x: float): float;
  begin
    result:= x;
    {
    if x < float(0.04045)
      then Result:= x * float(1 / 12.92)
      else Result:= power (x + float(0.055) / float (1 + 0.055), float(2.4));
  }
  end;

  function LinearTosRGB(x: float): float;
  begin
    Result:= x; {
    if x < float(0.0031308)
      then Result:= float(12.92) * x
      else Result:= float(1 + 0.055) * power(x, float(1 / 2.4)) - float(0.055);
}
  end;

  function ToGrayscale(const c: TColorFPRec): float;
  begin
    Result:= LinearTosRGB(
       float(0.2126) * sRGBToLinear(c.r)
     + float(0.7152) * sRGBToLinear(c.g)
     + float(0.0722) * sRGBToLinear(c.b));
  end;

  function TAbstractLayer.Blend(my, next: TColorFPRec): TColorFPRec;
    var
      r, g, b, s1, s2, v1, v2: float;
      h1, h2: TVector3f;
    function BlendFunc(a, b: float; mode: TLayerMode): float;
      function gw3c(a: float): float;
      begin
        if a < 0.25
          then Result:= ((16 * a - 12) * a + 4) * a
          else Result:= sqrt(a);
      end;
    begin
      case mode of
        lmMultiply: Result:= a * b;
        lmDivide:   Result:= b / max(0.0001, a);
        lmScreen:   Result:= 1 - (1 - b) * (1 - a);
        lmOverlay: begin
            if a < 0.5
              then Result:= 2 * a * b
              else Result:= 1 - 2 * (1 - a) * (1 - b);
          end;
        lmHard_Light: begin
            if b < 0.5
              then Result:= 2 * b * a
              else Result:= 1 - 2 * (1 - b) * (1 - a);
          end;
        lmSoft_Light: begin
            if  b < 0.5
              then Result:= a - (1 - 2 * b) * a * (1 - a)
              else Result:= a + (2 * b - 1) * (gw3c(a) - a);
          end;
        lmDodge:    Result:= b / max (0.0001, 1 - a);
        lmBurn:     Result:= 1 - ((1 - b) / max (0.0001, a));
        lmGrain_Extract: Result:= b - a + 0.5;
        lmGrain_Merge: Result:= a + b - 0.5;
        lmDifference: Result:= max(a, b) - min (a, b);
        lmAddition: Result:= a + b;
        lmSubstract: Result:= b - a;
        lmDarken_Only: Result:= min (a, b);
        lmLighten_Only: Result:= max (a, b);
      else
        skymp.Die('Error, the code does not support blend mode ' + GetEnumName(typeinfo(TLayerMode), Ord(mode)));
      end;
    end;
    procedure HSVtoRGB(const h: TVector3f; const s, v: float; out r, g, b: float);
    begin
      r:= v * (1 - s + h[0] * s);
      g:= v * (1 - s + h[1] * s);
      b:= v * (1 - s + h[2] * s);
    end;
    procedure RGBtoHSV(const c: TColorFPRec; out h: TVector3f; out s, v: float);
    var
      gray, chroma, d: float;
    begin
      gray:= min(c.g, min(c.r, c.b));
      h[0]:= c.r - gray;
      h[1]:= c.g - gray;
      h[2]:= c.b - gray;
      chroma:= max(h[0], max(h[1], h[2]));//sqrt(sqr(h[0]) + sqr(h[1]) + sqr(h[2]));
      if chroma < 0.0001 then begin
        h[0]:= float(1 / sqrt(3));
        h[1]:= h[0];
        h[2]:= h[0];
      end
      else begin
        d:= 1 / chroma;
        h[0]*= d;
        h[1]*= d;
        h[2]*= d;
      end;
      v:= gray + chroma;
      if chroma + gray < 0.0001
        then s:= 0
        else s:= chroma / (chroma + gray);
    end;

  begin
    my.a *= opacity;
    case mode of
      lmNormal: begin
        Result.a:= max (next.a, my.a);
        Result.r:= next.r * (1 - my.a) + my.r * my.a;
        Result.g:= next.g * (1 - my.a) + my.g * my.a;
        Result.b:= next.b * (1 - my.a) + my.b * my.a;
      end;
      lmHue, lmSaturation, lmColor, lmValue: begin
          RGBtoHSV(my, h1, s1, v1);
          RGBtoHSV(next, h2, s2, v2);
          case mode of
            lmHue: begin
                HSVtoRGB(h1, s2, v2, r, g, b);
              end;
            lmSaturation: begin
                HSVtoRGB(h2, s1, v2, r, g, b);
              end;
            lmColor: begin
                HSVtoRGB(h1, s1, v2, r, g, b);
              end;
            lmValue: begin
                HSVtoRGB(h2, s2, v1, r, g, b);
              end;
          end;
          Result.a:= next.a;
          Result.r:= r * my.a + next.r * (1 - my.a);
          Result.g:= g * my.a + next.g * (1 - my.a);
          Result.b:= b * my.a + next.b * (1 - my.a);
        end;
    else
      Result.a:= next.a;
      Result.r:= BlendFunc(my.r, next.r, mode) * my.a + next.r * (1 - my.a);
      Result.g:= BlendFunc(my.g, next.g, mode) * my.a + next.g * (1 - my.a);
      Result.b:= BlendFunc(my.b, next.b, mode) * my.a + next.b * (1 - my.a);
    end;
  end;

  operator * (a, b: TMatrix4f) c: TMatrix4f; inline;
  var i: integer;
  begin
    for i:= 0 to 3 do begin
      c[i,0]:= a[i,0] * b[0,0] + a[i,1] * b[1,0] + a[i,2] * b[2,0] + a[i,3] * b[3,0];
      c[i,1]:= a[i,0] * b[0,1] + a[i,1] * b[1,1] + a[i,2] * b[2,1] + a[i,3] * b[3,1];
      c[i,2]:= a[i,0] * b[0,2] + a[i,1] * b[1,2] + a[i,2] * b[2,2] + a[i,3] * b[3,2];
      c[i,3]:= a[i,0] * b[0,3] + a[i,1] * b[1,3] + a[i,2] * b[2,3] + a[i,3] * b[3,3];
    end;
  end;

  function CrossProduct(var a, b: TVector3f): TVector3f; inline;
  begin
    Result[0]:= a[1] * b[2] - a[2] * b[1];
    Result[1]:= a[2] * b[0] - a[0] * b[2];
    Result[2]:= a[0] * b[1] - a[1] * b[0]
  end;

  function DotProduct(var a, b: TVector3f): float; inline;
  begin
     Result:= a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
  end;

  operator * (a, b: TVector3f) c: TVector3f; inline;
  begin
    c[0]:= a[0] * b[0];
    c[1]:= a[1] * b[1];
    c[2]:= a[2] * b[2];
  end;

  operator + (a, b: TVector3f) c: TVector3f; inline;
  begin
    c[0]:= a[0] + b[0];
    c[1]:= a[1] + b[1];
    c[2]:= a[2] + b[2];
  end;

  operator * (a: TVector3f; b: float) c: TVector3f; inline;
  begin
    c[0]:= a[0] * b;
    c[1]:= a[1] * b;
    c[2]:= a[2] * b;
  end;

  operator - (a, b: TVector3f) c: TVector3f; inline;
  begin
    c[0]:= a[0] - b[0];
    c[1]:= a[1] - b[1];
    c[2]:= a[2] - b[2];
  end;

  operator * (a: TVector3f; b: TMatrix4f) c: TVector3f; inline;
  begin
    c[0]:= a[0] * b[0][0] + a[1] * b[0][1] + a[2] * b[0][2] + b[0][3];
    c[1]:= a[0] * b[1][0] + a[1] * b[1][1] + a[2] * b[1][2] + b[1][3];
    c[2]:= a[0] * b[2][0] + a[1] * b[2][1] + a[2] * b[2][2] + b[2][3];
  end;

  procedure Normalize(var v: TVector3f); inline;
  var f: float;
  begin
    f:= 1.0 / sqrt(sqr(v[0]) + sqr(v[1]) + sqr(v[2]));
    v[0]*= f;
    v[1]*= f;
    v[2]*= f;
  end;

  procedure Invert(const a: TMatrix4f; out b: TMatrix4f);
    type
      arfloat1 = array[1..999] of float;
      mypfloat = ^arfloat1;
    var
      t: integer;
    Procedure mdtgpd(n, rwidth: integer; Var al, ca: float; Var term: integer);
    Var
        posdef                               : boolean;
        i, j, k, kmin1, indk, indi           : integer;
        h, lkk, normr, normt, sumrowi, norma : float;
        pal, t                               : mypfloat;
    Begin
      If (n<1) Or (rwidth<1) Then
        Begin
          term := 3;
         exit
        End; {wrong input}
      getmem(t, sizeof(float)*n);
      pal := @al;
      normr := 0;
      posdef := true;
      norma := 0;
      For i:=1 To n Do
        Begin
          sumrowi := 0;
          For j:=1 To i Do
            sumrowi := sumrowi+abs(pal^[(i-1)*rwidth+j]);
          For j:=i+1 To n Do
            sumrowi := sumrowi+abs(pal^[(j-1)*rwidth+i]);
          If norma<sumrowi Then
            norma := sumrowi;
          t^[i] := 2*random-1;
         h := t^[i];
          h := abs(h);
          If normr<h Then
            normr := h
        End; {i}
      k := 0;
      while (k<n) and posdef Do
        Begin
          kmin1 := k;
         k := k+1;
          indk := (k-1)*rwidth;
          lkk := pal^[indk+k];
          For j:=1 To kmin1 Do
            lkk := lkk-sqr(pal^[indk+j]);
          If lkk <= 0 Then
            posdef := false
          Else
            Begin
              pal^[indk+k] := sqrt(lkk);
             lkk := pal^[indk+k];
              For i:=k+1 To n Do
                Begin
                  indi := (i-1)*rwidth;
                  h := pal^[indi+k];
                  For j:=1 To kmin1 Do
                    h := h-pal^[indk+j]*pal^[indi+j];
                  pal^[indi+k] := h/lkk
                End; {i}
              h := t^[k];
              For j:=1 To kmin1 Do
                h := h-pal^[indk+j]*t^[j];
              t^[k] := h/lkk
            End {posdef}
        End; {k}
      If posdef Then
        Begin
          normt := 0;
          For i:=n Downto 1 Do
            Begin
              h := t^[i];
              For j:=i+1 To n Do
                h := h-pal^[(j-1)*rwidth+i]*t^[j];
              t^[i] := h/pal^[(i-1)*rwidth+i];
              h := abs(t^[i]);
              If normt<h Then
                normt := h
            End; {i}
          ca := norma*normt/normr
        End; {posdef}
      If posdef Then
        term := 1
      Else
        term := 2;
      freemem(t);
    End;
    procedure invgpd(n, rwidth: integer; ai: pfloat; var term: integer);
    var success      : boolean;
        i, j, k, ind : integer;
        tk, h, ca    : float;
        pa, t        : mypfloat;
    begin
      if (n<1) or (rwidth<1) then
      begin
          term:=3; exit
      end; {wrong input}
      pa:= mypfloat(ai);
      mdtgpd(n, rwidth, pa^[1], ca, term);
      getmem(t, n*sizeof(float));
      if term=1 then
      begin
          success:=true; ind:=1; k:=1;
          while (k<>n+1) and success do
          begin
              success:=pa^[ind]<>0; k:=k+1; ind:=ind+rwidth+1
          end; {k}
          if success then
          begin
              for k:=n downto 1 do
              begin
                  for i:=k to n do t^[i]:=pa^[(i-1)*rwidth+k];
                  tk:=t^[k];
                  for i:=n downto k do
                  begin
                      if i=k then h:=1/tk else h:=0;
                      ind:=(i-1)*rwidth;
                      for j:=k+1 to i do h:=h-pa^[ind+j]*t^[j];
                      for j:=i+1 to n do h:=h-pa^[(j-1)*rwidth+i]*t^[j];
                      pa^[ind+k]:=h/tk
                  end {i}
              end {k}
          end; {success}
          if (not success) then term:=2 else
          for i:=1 to n do
          begin
              ind:=(i-1)*rwidth;
              for j:=i+1 to n do pa^[ind+j]:=pa^[(j-1)*rwidth+i]
          end; {i}
      end; {term=1}
      freemem(t);
    end; {invgpd}
  begin
    b:= a;
    invgpd(4, 4, @b[0][0], t);
  end;


  function TSkymp.ParseMode(_name: string): TLayerMode;
  var
    s, n, o: string;
    m: TLayerMode;
  begin
    if (_name = 'sky') or (_name = '') then Exit(lmNormal);
    s:= Trim(ReadString(_name, 'mode', ''));
    if s = '' then begin
      OutStr('  [' + _name + ']:mode is not specified, assuming normal');
      Exit(lmNormal);
    end;
    s:= LowerCase(s);
    for Result in TLayerMode do begin
      n:= GetEnumname(typeinfo(TLayerMode), ord(result));
      n:= LowerCase(COPY(n, 3, length(n) - 2));
      if n = s then Exit;
    end;

    //composing error message
    n:= '';
    for m in TLayerMode do begin
      if n <> '' then n+= ' ';
      o:= GetEnumname(typeinfo(TLayerMode), ord(m));
      n+= LowerCase(COPY(o, 3, length(o) - 2))
    end;
    Die('[' + _name + ']:mode has invalid value ' + s + '! Valid values are ' + n);

  end;

end.

